from __future__ import print_function, division
from coffea import processor

import multiprocessing

import warnings

import numpy as np

import sys
import pytest


def test_parsl_start_stop():
    parsl = pytest.importorskip("parsl", minversion="0.7.2")
    
    from coffea.processor.parsl.detail import (_parsl_initialize,
                                               _parsl_stop,
                                               _default_cfg)

    dfk = _parsl_initialize(config=_default_cfg)
    
    _parsl_stop(dfk)


def do_parsl_job(parsl_config):
    from coffea.processor.parsl.detail import (_parsl_initialize,
                                               _parsl_stop)
    from coffea.processor import run_parsl_job

    import os
    import os.path as osp
    
    filelist = {'ZJets': [osp.join(os.getcwd(),'tests/samples/nano_dy.root')],
                'Data'  : [osp.join(os.getcwd(),'tests/samples/nano_dimuon.root')]
    }
    treename='Events'

    from coffea.processor.test_items import NanoTestProcessor
    from coffea.processor.parsl.parsl_executor import parsl_executor
    
    dfk = _parsl_initialize(parsl_config)
    
    proc = NanoTestProcessor()
    
    hists = run_parsl_job(filelist, treename, processor_instance = proc, executor=parsl_executor, data_flow=dfk)
    
    hists2 = run_parsl_job(filelist, treename, processor_instance = proc, executor=parsl_executor, data_flow=dfk, executor_args={'flatten': False})
    
    _parsl_stop(dfk)
    
    assert( hists['cutflow']['ZJets_pt'] == 4 )
    assert( hists['cutflow']['ZJets_mass'] == 1 )
    assert( hists['cutflow']['Data_pt'] == 15 )
    assert( hists['cutflow']['Data_mass'] == 5 )
    
    assert( hists2['cutflow']['ZJets_pt'] == 4 )
    assert( hists2['cutflow']['ZJets_mass'] == 1 )
    assert( hists2['cutflow']['Data_pt'] == 15 )
    assert( hists2['cutflow']['Data_mass'] == 5 )

@pytest.mark.skipif(sys.platform.startswith('darwin'), reason='issue with parsl htex on macos')
def test_parsl_htex_executor():
    parsl = pytest.importorskip("parsl", minversion="0.7.2")

    from parsl.providers import LocalProvider
    from parsl.channels import LocalChannel
    from parsl.executors import HighThroughputExecutor
    from parsl.addresses import address_by_hostname
    from parsl.config import Config
    parsl_config = Config(
        executors=[
                   HighThroughputExecutor(
                    label="coffea_parsl_default",
                    address=address_by_hostname(),
                    cores_per_worker=max(multiprocessing.cpu_count()//2,1),
                    max_workers=1,
                    provider=LocalProvider(
                        channel=LocalChannel(),
                        init_blocks=1,
                        max_blocks=1,
                        nodes_per_block=1
                        ),
                    )
                   ],
        strategy=None,
        )

    do_parsl_job(parsl_config)


def test_parsl_funcs():
    parsl = pytest.importorskip("parsl", minversion="0.7.2")

    import os.path as osp
    from coffea.processor.parsl.detail import derive_chunks

    filename = osp.abspath('tests/samples/nano_dy.root')
    dataset = 'Z+Jets'
    treename = 'Events'
    chunksize = 20
    ds, tn, test = derive_chunks.func(filename, treename, chunksize, dataset)
    
    assert(dataset == ds)
    assert(treename == tn)
    assert('nano_dy.root' in test[0][0])
    assert(test[0][1] == 20)
    assert(test[0][2] == 0)

    from coffea.processor.parsl.parsl_executor import coffea_pyapp
    from coffea.processor.test_items import NanoTestProcessor
    import cloudpickle as cpkl
    import lz4.frame as lz4f
    
    procpkl = lz4f.compress(cpkl.dumps(NanoTestProcessor()))
    
    out = coffea_pyapp.func('test', filename, treename, chunksize, 0, procpkl)

    assert(len(out[0]) == 1348641)
    assert(out[1] == 10)
    assert(out[2] == 'test')

@pytest.mark.skipif(sys.platform.startswith('win'), reason='signals are different on windows')
def test_timeout():
    from coffea.processor.parsl.timeout import timeout
    import signal
    
    @timeout
    def too_long(timeout=None):
        import time
        time.sleep(20)

    @timeout
    def make_except(timeout=None):
        import time
        time.sleep(1)
        raise Exception('oops!')

    try:
        too_long(timeout=5)
    except Exception as e:
        assert(e.args[0] == "Timeout hit")

    try:
        make_except(timeout=20)
    except Exception as e:
        assert(e.args[0] == 'oops!')

    # reset alarms for other tests, this is suspicious
    signal.alarm(0)


def test_parsl_condor_cfg():
    parsl = pytest.importorskip("parsl", minversion="0.7.2")

    from coffea.processor.parsl.condor_config import condor_config

    test = condor_config()


def test_parsl_slurm_cfg():
    import os
    parsl = pytest.importorskip("parsl", minversion="0.7.2")

    x509_proxy = 'x509up_u%s' % (os.getuid())
    fname = '/tmp/%s' % x509_proxy
    with open(fname, 'w+'):
        os.utime(fname, None)
    
    from coffea.processor.parsl.slurm_config import slurm_config

    test = slurm_config()



from concurrent.futures import as_completed
import multiprocessing

from tqdm import tqdm

import parsl

from parsl.app.app import python_app

from parsl.providers import LocalProvider
from parsl.channels import LocalChannel
from parsl.config import Config
from parsl.executors import HighThroughputExecutor

from ..executor import futures_handler

from .timeout import timeout

try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

_default_cfg = Config(
    executors=[
        HighThroughputExecutor(
            label="coffea_parsl_default",
            cores_per_worker=1,
            provider=LocalProvider(
                channel=LocalChannel(),
                init_blocks=1,
                max_blocks=1,
            ),
        )
    ],
    strategy=None,
)


def _parsl_initialize(config=None):
    dfk = parsl.load(config)
    return dfk


def _parsl_stop(dfk):
    dfk.cleanup()
    parsl.clear()


@python_app
@timeout
def derive_chunks(filename, treename, chunksize, ds, timeout=None):
    import uproot
    from collections.abc import Sequence
    from concurrent.futures import ThreadPoolExecutor, TimeoutError

    uproot.XRootDSource.defaults["parallel"] = False

    afile = None
    for i in range(5):
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(uproot.open, filename)
            try:
                afile = future.result(timeout=5)
            except TimeoutError:
                afile = None
            else:
                break

    if afile is None:
        raise Exception('unable to open: %s' % filename)

    afile = uproot.open(filename)
    tree = None
    if isinstance(treename, str):
        tree = afile[treename]
    elif isinstance(treename, Sequence):
        for name in reversed(treename):
            if name in afile:
                tree = afile[name]
    else:
        raise Exception('treename must be a str or Sequence but is a %s!' % repr(type(treename)))

    if tree is None:
        raise Exception('No tree found, out of possible tree names: %s' % repr(treename))

    nentries = tree.numentries
    return ds, treename, [(filename, chunksize, index) for index in range(nentries // chunksize + 1)]


def _parsl_get_chunking(filelist, treename, chunksize, status=True, timeout=None):
    futures = set(derive_chunks(fn, treename, chunksize, ds, timeout=timeout) for ds, fn in filelist)

    items = []

    def chunk_accumulator(total, result):
        ds, treename, chunks = result
        for chunk in chunks:
            total.append((ds, chunk[0], treename, chunk[1], chunk[2]))

    futures_handler(futures, items, status, 'files', 'Preprocessing', futures_accumulator=chunk_accumulator)

    return items



from .lookup_base import lookup_base

from ..util import awkward
from ..util import numpy as np
from copy import deepcopy

from scipy.interpolate import interp1d


def masked_bin_eval(dim1_indices, dimN_bins, dimN_vals):
    dimN_indices = np.empty_like(dim1_indices)
    for i in np.unique(dim1_indices):
        idx = np.where(dim1_indices == i)
        dimN_indices[idx] = np.clip(np.searchsorted(dimN_bins[i],
                                                    dimN_vals[idx],
                                                    side='right') - 1,
                                    0, len(dimN_bins[i]) - 2)
    return dimN_indices


class jec_uncertainty_lookup(lookup_base):
    """
    This class defines a lookup table for jet energy scale uncertainties.

    The uncertainty values can be looked up with a call as follows:
        junc_lut = jec_uncertainty_lookup()
        uncertainties = junc_lut(JetProperty1=jet.property1,...)

    "uncertainties" will be of the same shape as the input jet properties.

    The list of required jet properties are given in junc_lut.signature
    """
    def __init__(self, formula, bins_and_orders, knots_and_vars):
        """
        The constructor takes the output of the "convert_junc_txt_file"
        text file converter, which returns a formula, bins, and an interpolation table.
        """
        super(jec_uncertainty_lookup, self).__init__()
        self._dim_order = bins_and_orders[1]
        self._bins = bins_and_orders[0]
        self._eval_vars = knots_and_vars[1]
        self._eval_knots = knots_and_vars[0]['knots']
        self._eval_downs = []
        self._eval_ups = []
        self._formula_str = formula.strip('"')
        self._formula = None
        if self._formula_str != 'None' and self._formula_str != '':
            raise Exception('jet energy uncertainties have no formula!')

        for binname in self._dim_order[1:]:
            binsaslists = self._bins[binname].tolist()
            self._bins[binname] = [np.array(bins) for bins in binsaslists]

        # convert downs and ups into interp1ds
        # (yes this only works for one binning dimension right now, fight me)
        for bin in range(self._bins[self._dim_order[0]].size - 1):
            self._eval_downs.append(interp1d(self._eval_knots,
                                             knots_and_vars[0]['downs'][bin]))
            self._eval_ups.append(interp1d(self._eval_knots,
                                           knots_and_vars[0]['ups'][bin]))

        # get the jit to compile if we've got more than one bin dim
        if len(self._dim_order) > 1:
            masked_bin_eval(np.array([0]), self._bins[self._dim_order[1]], np.array([0.0]))

        self._signature = deepcopy(self._dim_order)
        for eval in self._eval_vars:
            if eval not in self._signature:
                self._signature.append(eval)
        self._dim_args = {self._dim_order[i]: i for i in range(len(self._dim_order))}
        self._eval_args = {}
        for i, argname in enumerate(self._eval_vars):
            self._eval_args[argname] = i + len(self._dim_order)
            if argname in self._dim_args.keys():
                self._eval_args[argname] = self._dim_args[argname]

    def _evaluate(self, *args):
        """ uncertainties = f(args) """
        bin_vals = {argname: args[self._dim_args[argname]] for argname in self._dim_order}
        eval_vals = {argname: args[self._eval_args[argname]] for argname in self._eval_vars}

        # lookup the bins that we care about
        dim1_name = self._dim_order[0]
        dim1_indices = np.clip(np.searchsorted(self._bins[dim1_name],
                                               bin_vals[dim1_name],
                                               side='right') - 1,
                               0, self._bins[dim1_name].size - 2)

        # get clamp values and clip the inputs
        outs = np.ones(shape=(args[0].size, 2), dtype=np.float)
        for i in np.unique(dim1_indices):
            mask = np.where(dim1_indices == i)
            vals = np.clip(eval_vals[self._eval_vars[0]][mask],
                           self._eval_knots[0], self._eval_knots[-1])
            outs[:, 0][mask] += self._eval_ups[i](vals)
            outs[:, 1][mask] -= self._eval_downs[i](vals)

        return outs

    @property
    def signature(self):
        """ the list of all needed jet properties to be passed as kwargs to this lookup """
        return self._signature

    def __repr__(self):
        out = 'binned dims   : %s\n' % (self._dim_order)
        out += 'eval vars     : %s\n' % (self._eval_vars)
        out += 'signature     : (%s)\n' % (','.join(self._signature))
        return out



from __future__ import print_function, division
from coffea import processor
import numpy as np


def test_accumulators():
    a = processor.accumulator(0.)
    a += 3.
    a += processor.accumulator(2)
    assert a.value == 5.
    assert a.identity().value == 0.

    a = processor.accumulator(np.array([0.]))
    a += 3.
    a += processor.accumulator(2)
    assert a.value == np.array([5.])
    assert a.identity().value == np.array([0.])

    b = processor.set_accumulator({'apples', 'oranges'})
    b += {'pears'}
    b += 'grapes'
    assert b == {'apples', 'oranges', 'pears', 'grapes'}

    c = processor.dict_accumulator({'num': a, 'fruit': b})
    c['num'] += 2.
    c += processor.dict_accumulator({
        'num2': processor.accumulator(0),
        'fruit': processor.set_accumulator({'apples', 'cherries'}),
    })
    assert c['num2'].value == 0
    assert c['num'].value == 7.
    assert c['fruit'] == {'apples', 'oranges', 'pears', 'grapes', 'cherries'}

    d = processor.defaultdict_accumulator(lambda: processor.accumulator(0.))
    d['x'] = processor.accumulator(0.)
    d['x'] += 4.
    d['y'] += 5.
    d['z'] += d['x']
    d['x'] += d['y']
    assert d['x'].value == 9.
    assert d['y'].value == 5.
    assert d['z'].value == 4.
    assert d['w'].value == 0.

    e = d + c

    f = processor.defaultdict_accumulator(lambda: 2.)
    f['x'] += 4.
    assert f['x'] == 6.

    f += f
    assert f['x'] == 12.
    assert f['y'] == 2.



from __future__ import print_function, division

from coffea.striped import ColumnGroup, PhysicalColumnGroup, jaggedFromColumnGroup
import uproot
import uproot_methods
from coffea.util import awkward
from coffea.util import numpy as np

from dummy_distributions import dummy_events

def test_striped():
    events = dummy_events()
    colgroup = ColumnGroup(events,"thing","p4","blah")
    physcolgroup = PhysicalColumnGroup(events,"thing","blah",p4="p4")
    physcolgroup2 = PhysicalColumnGroup(events,"thing","blah",
                                        pt="pt",eta="eta",phi="phi",mass="mass")
    physcolgroup3 = PhysicalColumnGroup(events,"thing","blah",
                                        px="px",py="py",pz="pz",energy="en")
    
    jagged = jaggedFromColumnGroup(colgroup)
    candjagged = jaggedFromColumnGroup(physcolgroup)
    candjagged2 = jaggedFromColumnGroup(physcolgroup2)
    candjagged3 = jaggedFromColumnGroup(physcolgroup3)
    
    assert not hasattr(jagged,'p4')
    assert hasattr(candjagged,'p4')
    assert hasattr(candjagged2,'p4')
    assert hasattr(candjagged3,'p4')



from coffea.util import load, save
from coffea.processor.test_items import NanoTestProcessor
import os

def test_loadsave():
    filename = 'testprocessor.coffea'
    try:
        aprocessor = NanoTestProcessor()
        save(aprocessor, filename)
        newprocessor = load(filename)
        assert 'pt' in newprocessor.accumulator
        assert newprocessor.accumulator['pt'].compatible(aprocessor.accumulator['pt'])
    finally:
        if os.path.exists(filename):
            os.remove(filename)



# here we want to define a group of columns from striped
# that all pertain to the same object type
# i.e. group = PhysicalColumnGroup(job,p4="Electron_p4",""NElectrons","Electron_p4","Electron_id1",...)
# this is specialized to striped so far
# we should try to peel it apart for other things


class ColumnGroup(object):
    def __init__(self, events, objName, *args):
        self._map = {}
        eventObj = getattr(events, objName)
        self._counts = getattr(events, objName).count
        for arg in args:
            callStack = arg.split('.')
            retval = getattr(eventObj, callStack[0])
            for i in range(1, len(callStack)):
                retval = getattr(retval, callStack[i])
            self._map[arg] = retval

    def __getitem__(self, name):
        return self._map[name]

    def columnsWithout(self, toremove):
        out = {}
        out.update(self._map)
        if isinstance(toremove, str):
            del out[toremove]
        else:
            for key in toremove:
                del out[key]
        return out

    def columns(self):
        return self._map

    def counts(self):
        return self._counts


class PhysicalColumnGroup(ColumnGroup):
    def __init__(self, events, objName, *args, **kwargs):
        self._hasp4 = False
        argkeys = kwargs.keys()
        self._p4 = None
        if 'p4' in argkeys:
            self._hasp4 = True
            self._p4 = {'p4': kwargs['p4']}
        elif 'pt' in argkeys and 'eta' in argkeys and 'phi' in argkeys and 'mass' in argkeys:
            self._hasp4 = True
            self._p4 = {key: kwargs[key] for key in ['pt', 'eta', 'phi', 'mass']}
        elif 'pt' in argkeys and 'eta' in argkeys and 'phi' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key: kwargs[key] for key in ['pt', 'eta', 'phi', 'energy']}
        elif 'px' in argkeys and 'py' in argkeys and 'pz' in argkeys and 'mass' in argkeys:
            self._hasp4 = True
            self._p4 = {key: kwargs[key] for key in ['px', 'py', 'pz', 'mass']}
        elif 'pt' in argkeys and 'phi' in argkeys and 'pz' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key: kwargs[key] for key in ['pt', 'phi', 'pz', 'energy']}
        elif 'px' in argkeys and 'py' in argkeys and 'pz' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key: kwargs[key] for key in ['px', 'py', 'pz', 'energy']}
        elif 'p' in argkeys and 'theta' in argkeys and 'phi' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key: kwargs[key] for key in ['p', 'theta', 'phi', 'energy']}
        elif 'p3' in argkeys and 'energy' in argkeys:
            self._hasp4 = True
            self._p4 = {key: kwargs[key] for key in ['p3', 'energy']}
        else:
            raise Exception('No valid definition of four-momentum found to build JaggedCandidateArray')
        allargs = list(args) + [val for val in kwargs.values()]
        super(PhysicalColumnGroup, self).__init__(events, objName, *allargs)

    def p4Name(self):
        if self._p4 is None:
            raise Exception('p4 is not set for this PhysicalColumnGroup')
        return self._p4.values()

    def p4Columns(self):
        return {key: self.columns()[value] for key, value in self._p4.items()}

    def otherColumns(self):
        return self.columnsWithout(self.p4Name())



from .processor import ProcessorABC
from .dataframe import (
    LazyDataFrame,
    PreloadedDataFrame,
)
from .helpers import Weights, PackedSelection
from .executor import (
    iterative_executor,
    futures_executor,
    run_uproot_job,
    run_parsl_job,
    run_spark_job
)
from .accumulator import (
    accumulator,
    set_accumulator,
    dict_accumulator,
    defaultdict_accumulator,
)
